#include "description_parser.h"
#include "processor.h"
#include "shader_compiler.h"

#include "runtime/buffer_manager.h"
#include "runtime/config.h"
#include "runtime/gfx.h"
#include "runtime/mem_arena.h"
#include "runtime/runtime.h"

#include <assert.h>
#include <limits.h>
#include <stdbool.h>
#include <string.h>

typedef struct {
    rt_attribute_binding *uniform_bindings;
    rt_attribute_binding *storage_bindings;
    rt_attribute_binding *texture_bindings;

    rt_resource shaders[3];
    char *shader_names[3];
    unsigned int shader_count;

    unsigned int vertex_shader;
    unsigned int fragment_shader;
    unsigned int compute_shader;

    /* TODO Fixed function settings */

    /* Sampler settings */

    uint16_t uniform_binding_count;
    uint16_t storage_binding_count;
    uint16_t texture_binding_count;
} rt_parsed_pipeline_data;

enum {
    RT_SHADER_NOT_PRESENT = RT_ASSET_PROCESSING_FAILED + 1
};

extern RT_DLLIMPORT rt_cvar rt_Renderer;

static bool ParseBindingIndex(rt_text_span span, unsigned int *index) {
    if (span.length == 0)
        return false;
    int at           = (int)span.length - 1;
    unsigned int exp = 1;
    unsigned int n   = 0;
    while (at >= 0) {
        if (span.start[at] >= '0' && span.start[at] <= '9') {
            unsigned int digit = (unsigned int)(span.start[at] - '0');
            n += digit * exp;
        } else {
            rtReportError("GFX", "Unexpected non-digit character in binding index");
            return false;
        }
        --at;
        exp *= 10;
    }
    *index = n;
    return true;
}

static rt_attribute_value ParseBindingValue(rt_text_span span) {
    if (rtCompareSpanToString(span, "MATERIAL_ALBEDO") == 0) {
        return RT_ATTRIBUTE_VALUE_MATERIAL_ALBEDO;
    } else if (rtCompareSpanToString(span, "MATERIAL_NORMAL") == 0) {
        return RT_ATTRIBUTE_VALUE_MATERIAL_NORMAL;
    }
    rtReportError("GFX", "Unsupported binding value %*.s", span.length, span.start);
    return RT_ATTRIBUTE_VALUE_UNDEFINED;
}

static bool ParseBindings(rt_parse_state *state,
                          unsigned int root_list,
                          const char *name,
                          const char *file_path,
                          rt_attribute_binding **p_bindings,
                          uint16_t *p_binding_count) {
    const rt_parsed_stmt *bindings = rtFindStatement(state, root_list, name);
    if (bindings) {
        if (bindings->form != RT_STMT_FORM_LIST) {
            rtReportError("GFX",
                          "Expected list of bindings as the value of "
                          "\"%s\" in %s",
                          name,
                          file_path);
            return false;
        }
        const rt_parsed_stmt_list *binding_list = &state->statement_lists[bindings->list_index];
        rt_attribute_binding *shader_bindings =
            rtAllocBuffer(sizeof(rt_attribute_binding) * binding_list->count);
        if (!bindings) {
            rtReportError("GFX", "Out of memory");
            return false;
        }
        unsigned int binding_count = binding_list->count;

        unsigned int stmt_index = binding_list->first;
        for (unsigned int i = 0; i < binding_list->count; ++i) {
            const rt_parsed_stmt *stmt = &state->statements[stmt_index];
            if (!ParseBindingIndex(stmt->attribute, &shader_bindings[i].index)) {
                rtReleaseBuffer(shader_bindings,
                                sizeof(rt_attribute_binding) * binding_list->count);
                return false;
            }

            shader_bindings[i].value = ParseBindingValue(stmt->value);
            if (shader_bindings[i].value == RT_ATTRIBUTE_VALUE_UNDEFINED) {
                rtReleaseBuffer(shader_bindings,
                                sizeof(rt_attribute_binding) * binding_list->count);
                return false;
            }
            stmt_index = stmt->next;
        }

        *p_bindings      = shader_bindings;
        *p_binding_count = (uint16_t)binding_count;
        return true;
    } else {
        *p_bindings      = NULL;
        *p_binding_count = 0;
        return true;
    }
}

static char *GenerateShaderName(rt_shader_type type,
                                rt_shader_stage stage,
                                rt_shader_optimization_level optimization,
                                const char *file_name,
                                rt_arena *arena) {
    size_t name_len = strlen(file_name) + 5 /* type */
                      + 5                   /* stage */
                      + 3                   /* optimization */
                      + 1;                  /* '\0' */
    char *res_name = rtArenaPush(arena, name_len);
    if (!res_name)
        return NULL;
    const char *type_str = NULL;
    switch (type) {
    case RT_SHADER_TYPE_VULKAN:
        type_str = ":vk";
        break;
    default:
        return NULL;
    }
    const char *stage_str = NULL;
    switch (stage) {
    case RT_SHADER_STAGE_VERTEX:
        stage_str = ":vert";
        break;
    case RT_SHADER_STAGE_FRAGMENT:
        stage_str = ":frag";
        break;
    case RT_SHADER_STAGE_COMPUTE:
        stage_str = ":comp";
        break;
    default:
        return NULL;
    }
    const char *optim_str = NULL;
    switch (optimization) {
    case RT_SHADER_OPTIMIZATION_NONE:
        optim_str = ":O0";
        break;
    case RT_SHADER_OPTIMIZATION_SIZE:
        optim_str = ":Os";
        break;
    case RT_SHADER_OPTIMIZATION_SPEED:
        optim_str = ":Ox";
        break;
    default:
        return NULL;
    }
    rtSPrint(res_name, name_len, "%s%s%s%s", file_name, type_str, stage_str, optim_str);
    return res_name;
}

static rt_result ParseShader(rt_parse_state *state,
                             unsigned int root_list,
                             const char *name,
                             const char *file_path,
                             rt_shader_type type,
                             rt_shader_stage stage,
                             rt_shader_optimization_level optimization,
                             rt_resource *p_resource,
                             char **p_resource_name,
                             rt_arena *arena) {
    const rt_parsed_stmt *stmt = rtFindStatement(state, root_list, name);
    if (stmt) {
        if (stmt->form != RT_STMT_FORM_LIST) {
            rtReportError("GFX",
                          "Expected a list as the value of "
                          "\"%s\" in %s",
                          name,
                          file_path);
            return RT_ASSET_PROCESSING_FAILED;
        }
        const rt_parsed_stmt_list *shader_list = &state->statement_lists[stmt->list_index];

        unsigned int stmt_index = shader_list->first;
        for (unsigned int i = 0; i < shader_list->count; ++i) {
            const rt_parsed_stmt *shader = &state->statements[stmt_index];
            if (shader->form != RT_STMT_FORM_VALUE && shader->form != RT_STMT_FORM_BLOCK) {
                rtReportError("GFX",
                              "Expected a simple statement or block for "
                              "\"%s.%*s\" in %s",
                              name,
                              (int)shader->attribute.length,
                              shader->attribute.start,
                              file_path);
                return RT_ASSET_PROCESSING_FAILED;
            }
            rt_shader_type in_file_type = RT_SHADER_TYPE_INVALID;
            if (rtCompareSpanToString(shader->attribute, "vk") == 0) {
                in_file_type = RT_SHADER_TYPE_VULKAN;
            } else {
                rtReportError("GFX",
                              "Invalid renderer backend"
                              "\"%*s\" in %s of file %s",
                              (int)shader->attribute.length,
                              shader->attribute.start,
                              name,
                              file_path);
                return RT_ASSET_PROCESSING_FAILED;
            }

            if (in_file_type == type) {
                if (shader->form == RT_STMT_FORM_BLOCK) {
                    /* Inline code */
                    rt_shader_bytecode bytecode =
                        CompileShader(type, stage, optimization, shader->block, file_path, arena);
                    if (!bytecode.bytes)
                        return RT_ASSET_PROCESSING_FAILED;

                    *p_resource_name =
                        GenerateShaderName(type, stage, optimization, file_path, arena);
                    if (!*p_resource_name)
                        return RT_ASSET_PROCESSING_FAILED;

                    rt_resource resource;
                    resource.type              = RT_RESOURCE_SHADER;
                    resource.dependency_count  = 0;
                    resource.subresource_count = 0;
                    resource.data = rtArenaPush(arena, sizeof(rt_shader_info) + bytecode.len);
                    if (!resource.data)
                        return RT_ASSET_PROCESSING_FAILED;
                    rt_shader_info *shader_info  = resource.data;
                    uint8_t *shader_bytecode     = (uint8_t *)(shader_info + 1);
                    shader_info->stage           = stage;
                    shader_info->type            = type;
                    shader_info->bytecode_length = bytecode.len;
                    rtSetRelptr(&shader_info->bytecode, shader_bytecode);
                    memcpy(shader_bytecode, bytecode.bytes, bytecode.len);

                    memcpy(p_resource, &resource, sizeof(resource));
                    break;
                } else if (shader->form != RT_STMT_FORM_VALUE) {
                    /* A filename */
                    rtLog("AC", "Shader source files not implemented yet!");
                    return RT_ASSET_PROCESSING_FAILED;
                }
            }

            stmt_index = shader->next;
        }
        return RT_SUCCESS;
    }
    return RT_SHADER_NOT_PRESENT;
}

static rt_shader_optimization_level
ParseOptimizationLevel(rt_parse_state *state, unsigned int root_list, const char *file_path) {
    const rt_parsed_stmt *stmt = rtFindStatement(state, root_list, "optimization");
    rt_shader_optimization_level optimization_level = RT_SHADER_OPTIMIZATION_NONE;
    if (stmt) {
        if (stmt->form != RT_STMT_FORM_VALUE) {
            rtReportError("GFX",
                          "Expected a simple statement for"
                          "\"optimization\" in %s",
                          file_path);
            return RT_SHADER_OPTIMIZATION_NONE;
        }

        if (rtCompareSpanToString(stmt->value, "speed") == 0) {
            optimization_level = RT_SHADER_OPTIMIZATION_SPEED;
        } else if (rtCompareSpanToString(stmt->value, "size") == 0) {
            optimization_level = RT_SHADER_OPTIMIZATION_SIZE;
        } else if (rtCompareSpanToString(stmt->value, "none") == 0) {
            optimization_level = RT_SHADER_OPTIMIZATION_NONE;
        } else {
            rtReportError("GFX",
                          "Expected one of 'speed', 'size' and 'none' for \"optimization\" in %s",
                          file_path);
        }
    }

    return optimization_level;
}

static rt_result ParsePipelineFile(rt_file_id fid,
                                   const char *text,
                                   size_t length,
                                   rt_parsed_pipeline_data *pipeline,
                                   rt_arena *arena) {
    /* This is the grammar for pipeline files:
     * <stmt-list> ::= <stmt>*
     * <stmt> ::= <attribute> ( ( <value> ';' ) | ( '{' <stmt-list> '}' ) )
     * <attribute> ::= [:alnum:]*
     * <value>:: = [:alnum:]* */
    const char *file_path = rtGetFilePath(fid);
    rt_parse_state state;
    unsigned int root_list;
    rt_result result = rtParseDescription(text, length, file_path, &root_list, &state, arena);
    if (result != RT_SUCCESS) {
        goto out;
    }

    /* We allow the pipeline file to overwrite the optimization level */
    rt_shader_optimization_level optimization =
        ParseOptimizationLevel(&state, root_list, file_path);

    rt_shader_type type = RT_SHADER_TYPE_INVALID;
    if (strcmp(rt_Renderer.s, "vk") == 0)
        type = RT_SHADER_TYPE_VULKAN;

    if (type == RT_SHADER_TYPE_INVALID) {
        result = RT_ASSET_PROCESSING_FAILED;
        rtLog("AC", "rt_Renderer (%s) could not be translated to a shader type.", rt_Renderer.s);
        goto out;
    }

    /* Process shader stages */
    result = ParseShader(&state,
                         root_list,
                         "vertex",
                         file_path,
                         type,
                         RT_SHADER_STAGE_VERTEX,
                         optimization,
                         &pipeline->shaders[pipeline->shader_count],
                         &pipeline->shader_names[pipeline->shader_count],
                         arena);
    if (result == RT_SUCCESS) {
        pipeline->vertex_shader = pipeline->shader_count;
        ++pipeline->shader_count;
    } else if (result == RT_SHADER_NOT_PRESENT) {
        pipeline->vertex_shader = UINT_MAX;
    } else {
        goto out;
    }
    result = RT_SUCCESS;

    result = ParseShader(&state,
                         root_list,
                         "fragment",
                         file_path,
                         type,
                         RT_SHADER_STAGE_FRAGMENT,
                         optimization,
                         &pipeline->shaders[pipeline->shader_count],
                         &pipeline->shader_names[pipeline->shader_count],
                         arena);
    if (result == RT_SUCCESS) {
        pipeline->fragment_shader = pipeline->shader_count;
        ++pipeline->shader_count;
    } else if (result == RT_SHADER_NOT_PRESENT) {
        pipeline->fragment_shader = UINT_MAX;
    } else {
        goto out;
    }
    result = RT_SUCCESS;

    result = ParseShader(&state,
                         root_list,
                         "compute",
                         file_path,
                         type,
                         RT_SHADER_STAGE_COMPUTE,
                         optimization,
                         &pipeline->shaders[pipeline->shader_count],
                         &pipeline->shader_names[pipeline->shader_count],
                         arena);
    if (result == RT_SUCCESS) {
        pipeline->compute_shader = pipeline->shader_count;
        ++pipeline->shader_count;
    } else if (result == RT_SHADER_NOT_PRESENT) {
        pipeline->compute_shader = UINT_MAX;
    } else {
        goto out;
    }
    result = RT_SUCCESS;

    /* Process bindings */
    pipeline->texture_bindings      = NULL;
    pipeline->texture_binding_count = 0;
    pipeline->uniform_bindings      = NULL;
    pipeline->uniform_binding_count = 0;
    pipeline->storage_bindings      = NULL;
    pipeline->storage_binding_count = 0;

    if (!ParseBindings(&state,
                       root_list,
                       "texture_bindings",
                       file_path,
                       &pipeline->texture_bindings,
                       &pipeline->texture_binding_count)) {
        result = RT_ASSET_PROCESSING_FAILED;
        goto out;
    }

    if (!ParseBindings(&state,
                       root_list,
                       "uniform_bindings",
                       file_path,
                       &pipeline->uniform_bindings,
                       &pipeline->uniform_binding_count)) {
        result = RT_ASSET_PROCESSING_FAILED;
        goto out;
    }

    if (!ParseBindings(&state,
                       root_list,
                       "storage_bindings",
                       file_path,
                       &pipeline->storage_bindings,
                       &pipeline->storage_binding_count)) {
        result = RT_ASSET_PROCESSING_FAILED;
        goto out;
    }

out:
    return result;
}

RT_ASSET_PROCESSOR_FN(PipelineProcessor) {
    rt_loaded_asset asset = LoadAsset(file);
    if (!asset.buffer)
        return RT_UNKNOWN_ERROR;

    rt_parsed_pipeline_data pipeline;
    memset(&pipeline, 0, sizeof(pipeline));
    rt_result result = ParsePipelineFile(file, asset.buffer, asset.size, &pipeline, arena);
    if (result != RT_SUCCESS)
        goto out;

    rt_resource_id shader_resources[3] = {0};
    result                             = rtCreateResources(pipeline.shader_count,
                               pipeline.shader_names,
                               pipeline.shaders,
                               shader_resources);
    if (result != RT_SUCCESS)
        goto out;

    rt_resource pipeline_resource      = {0};
    pipeline_resource.type             = RT_RESOURCE_PIPELINE;
    pipeline_resource.dependency_count = pipeline.shader_count;
    memcpy(pipeline_resource.dependencies, shader_resources, sizeof(shader_resources));
    pipeline_resource.subresource_count = 0;

    size_t data_size =
        sizeof(rt_pipeline_info) + sizeof(rt_attribute_binding) * (pipeline.texture_binding_count +
                                                                   pipeline.uniform_binding_count +
                                                                   pipeline.storage_binding_count);
    pipeline_resource.data = rtArenaPush(arena, data_size);
    if (!pipeline_resource.data) {
        result = RT_OUT_OF_MEMORY;
        goto out;
    }
    rt_pipeline_info *info = pipeline_resource.data;
    memset(info, 0, sizeof(*info));
    info->vertex_shader                    = (pipeline.vertex_shader != UINT_MAX)
                                                 ? shader_resources[pipeline.vertex_shader]
                                                 : RT_INVALID_RESOURCE_ID;
    info->fragment_shader                  = (pipeline.fragment_shader != UINT_MAX)
                                                 ? shader_resources[pipeline.fragment_shader]
                                                 : RT_INVALID_RESOURCE_ID;
    info->compute_shader                   = (pipeline.compute_shader != UINT_MAX)
                                                 ? shader_resources[pipeline.compute_shader]
                                                 : RT_INVALID_RESOURCE_ID;
    rt_attribute_binding *uniform_bindings = (rt_attribute_binding *)(info + 1);
    if (pipeline.uniform_binding_count > 0) {
        memcpy(uniform_bindings, pipeline.uniform_bindings, pipeline.uniform_binding_count);
        rtSetRelptr(&info->uniform_bindings, uniform_bindings);
    }
    rt_attribute_binding *texture_bindings = (uniform_bindings + pipeline.uniform_binding_count);
    if (pipeline.texture_binding_count > 0) {
        memcpy(texture_bindings, pipeline.texture_bindings, pipeline.texture_binding_count);
        rtSetRelptr(&info->texture_bindings, texture_bindings);
    }
    rt_attribute_binding *storage_bindings = (texture_bindings + pipeline.texture_binding_count);
    if (pipeline.texture_binding_count > 0) {
        memcpy(storage_bindings, pipeline.storage_bindings, pipeline.storage_binding_count);
        rtSetRelptr(&info->storage_bindings, storage_bindings);
    }
    rt_resource_id pipeline_id;
    const char *name = rtGetFilePath(file);
    result           = rtCreateResources(1, &name, &pipeline_resource, &pipeline_id);
    if (result == RT_SUCCESS) {
        new_resources[0] = pipeline_id;
        memcpy(&new_resources[1], shader_resources, sizeof(shader_resources));
        *new_resource_count = 1 + pipeline.shader_count;
    }
out:
    rtReleaseBuffer(asset.buffer, asset.size);
    return result;
}
